/*
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 *
 */

package com.openexchange.office.tools.database;

import static org.apache.commons.lang.StringUtils.isEmpty;
import static org.apache.commons.lang.StringUtils.isNotEmpty;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.openexchange.office.tools.annotation.NonNull;
import com.zaxxer.hikari.HikariConfig;
import com.zaxxer.hikari.pool.HikariPool;

import liquibase.Liquibase;
import liquibase.changelog.ChangeSet;
import liquibase.changelog.ChangeSet.ExecType;
import liquibase.changelog.ChangeSet.RunStatus;
import liquibase.changelog.DatabaseChangeLog;
import liquibase.changelog.visitor.ChangeExecListener;
import liquibase.database.Database;
import liquibase.database.jvm.JdbcConnection;
import liquibase.exception.LiquibaseException;

/**
 * {@link DocumentsDatabase}
 *
 * @author <a href="mailto:kai.ahrens@open-xchange.com">Kai Ahrens</a>
 * @since v7.10.0
 */
public class DocumentsDatabase implements ChangeExecListener {

    /**
     * Initializes a new {@link DocumentsDatabase}.
     */
    public DocumentsDatabase(
        final DatabaseResourceAccessor resourceAccessor,
        final ConnectionData readConnectionData,
        final ConnectionData writeConnectionData,
        boolean updateSchema) throws DatabaseException {

        this.resourceAccessor = resourceAccessor;

        this.readConnectionData = readConnectionData;
        this.writeConnectionData = (null != writeConnectionData) ? writeConnectionData : readConnectionData;


        if (isEmpty(this.readConnectionData.getConnectionURL()) ) {

            throw new DatabaseException("Property readUrl must not be empty (Example Url: jdbc:mysql://host:port/databasename)");
        }

        // create connection pools
        implCreateConnectionPools(readConnectionData != null, writeConnectionData != null);

        // perform Liquibase update
        if (updateSchema) {
        	implUpdateDatabase();
        }
    }

    public DocumentsDatabase() {    	
    }
    
    protected void setReadConnectionData(ConnectionData readConnectionData) {
		this.readConnectionData = readConnectionData;
	}

	/*
     * (non-Javadoc)
     *
     * @see liquibase.changelog.visitor.ChangeExecListener#willRun(liquibase.changelog.ChangeSet, liquibase.changelog.DatabaseChangeLog, liquibase.database.Database, liquibase.changelog.ChangeSet.RunStatus)
     */
    @Override
    public void willRun(ChangeSet changeSet, DatabaseChangeLog databaseChangeLog, Database database, RunStatus runStatus) {
        if (null != changeSet) {
            implLogChangeEvent(runStatus, changeSet.getDescription());
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see liquibase.changelog.visitor.ChangeExecListener#ran(liquibase.changelog.ChangeSet, liquibase.changelog.DatabaseChangeLog, liquibase.database.Database, liquibase.changelog.ChangeSet.ExecType)
     */
    @Override
    public void ran(ChangeSet changeSet, DatabaseChangeLog databaseChangeLog, Database database, ExecType execType) {
        if (null != changeSet) {
            implLogChangeEvent(execType, changeSet.getDescription());
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see liquibase.changelog.visitor.ChangeExecListener#rolledBack(liquibase.changelog.ChangeSet, liquibase.changelog.DatabaseChangeLog, liquibase.database.Database)
     */
    @Override
    public void rolledBack(ChangeSet changeSet, DatabaseChangeLog databaseChangeLog, Database database) {
        if (null != changeSet) {
            implLogChangeEvent(null, changeSet.getDescription());
        }
    }

    // - Database access methods -----------------------------------------------

    /**
     * getReadConnection
     */
    public Connection getReadConnection() throws DatabaseException {
        if (null == this.readConnectionPool) {
            throw new DatabaseException("No read connection pool available. Please check logging output and setup!");
        }

        try {
            return this.readConnectionPool.getConnection();
        } catch (SQLException e) {
            throw new DatabaseException("Exception caught when getting read connection from connectiin pool", e);
        }
    }

    //  - Implementation -------------------------------------------------------

    /**
     * getWriteConnection
     */
    public Connection getWriteConnection() throws DatabaseException {
        if (null == this.writeConnectionPool) {
            throw new DatabaseException("No write connection pool available. Please check logging output and setup!");
        }

        try {
            return this.writeConnectionPool.getConnection();
        } catch (SQLException e) {
            throw new DatabaseException("Exception caught when getting write connection from connection pool", e);
        }
    }

    /**
     * @throws DatabaseException
     * @throws LiquibaseException
     */
    private void implUpdateDatabase() throws DatabaseException {
        final String resourceName = this.resourceAccessor.getResourceName();

        try (final Connection con = getWriteConnection()) {
            final Liquibase liquibase = new Liquibase(resourceName, this.resourceAccessor, new JdbcConnection(con));

            liquibase.setChangeExecListener(this);
            liquibase.update(null);
        } catch (Exception e) {
            throw new DatabaseException("Exception caught when updating database via Liquibase resource: " + resourceName, e);
        }
    }

    /**
     * @throws DatabaseException
     */
    protected void implCreateConnectionPools(boolean createReadConn, boolean createWriteConn) throws DatabaseException {
    	if (readConnectionData != null) {
    		this.readConnectionPool = implCreateConnectionPool(this.readConnectionData, this.readConnectionData, true);
    	}
    	if (writeConnectionData != null) {
    		this.writeConnectionPool = implCreateConnectionPool(this.writeConnectionData, this.writeConnectionData, false);
    	}
    }

    /**
     * @param connectionData
     * @param poolData
     * @param readOnly
     * @return
     * @throws DatabaseException
     */
    private static HikariPool implCreateConnectionPool(@NonNull final DatabaseConnectionData connectionData,
        @NonNull final ConnectionPoolData poolData, boolean readOnly) throws DatabaseException {

        final HikariConfig poolConfig = new HikariConfig();
        final Properties jdbcDatabaseProperties = connectionData.getJDBCProperties();
        final String driverClassName = connectionData.getDriverClassName();

        try {
            Class.forName(driverClassName);
            poolConfig.setDriverClassName(driverClassName);
        } catch (ClassNotFoundException e) {
            throw new DatabaseException("Exception caught when getting JDBC driver runtime class: " + driverClassName, e);
        }

        // DB related Hikari properties
        poolConfig.setJdbcUrl(connectionData.getConnectionURL());
        poolConfig.setUsername(connectionData.getUserName());
        poolConfig.setPassword(connectionData.getPassword());
        poolConfig.setAutoCommit(false);
        poolConfig.setReadOnly(readOnly);

        // JDBC driver properties
        if (jdbcDatabaseProperties != null) {
	        for (final String curPropKey : jdbcDatabaseProperties.stringPropertyNames()) {
	            poolConfig.addDataSourceProperty(curPropKey, jdbcDatabaseProperties.getProperty(curPropKey));
	        }
        }

        // Hikari pool properties
        poolConfig.setMaximumPoolSize(poolData.getConnectionPoolSize());
        poolConfig.setConnectionTimeout(poolData.getConnectionPoolConnectTimeoutMillis());
        poolConfig.setIdleTimeout(poolData.getConnectionPoolIdleTimeoutMillis());

        HikariPool result = null;
        try {
            result = new HikariPool(poolConfig);
        } catch (final Exception e) {
            final Map<String, String> props = new HashMap<>();
            props.put("driverClassName", poolConfig.getDriverClassName());
            props.put("jdbcUrl", poolConfig.getJdbcUrl());
            props.put("username", poolConfig.getUsername());
            props.put("password", StringUtils.isNotEmpty(poolConfig.getPassword()) ? "<set>" : "<empty>");
            throw new DatabaseException("Exception caught trying to create database connection pool", e, props);
        }

        return result;
    }

    /**
     * @param status
     * @param description
     */
    private static void implLogChangeEvent(final Object status, final String description) {
        if (LOG.isInfoEnabled()) {
            if (null != status) {
                LOG.debug("FileItemService database update received status: " + status.toString());
            }

            if (isNotEmpty(description)) {
                LOG.debug("FileItemService database update description: " + description);
            }
        }
    }

    // - Static members --------------------------------------------------------

    final private static Logger LOG = LoggerFactory.getLogger(DocumentsDatabase.class);

    // - Members ---------------------------------------------------------------

    private DatabaseResourceAccessor resourceAccessor;

    private ConnectionData readConnectionData;

    private ConnectionData writeConnectionData;

    private HikariPool readConnectionPool = null;

    private HikariPool writeConnectionPool = null;
}
